@@ -26,9 +26,9 @@
 extern PyObject *BasicAuthException_class;
 static void set_basicauth_error(krb5_context context, krb5_error_code code);
 
-static krb5_error_code verify_krb5_user(krb5_context context, krb5_principal principal, const char *password, krb5_principal server);
+static krb5_error_code verify_krb5_user(krb5_context context, krb5_principal principal, const char *password, krb5_principal server, unsigned char verify);
 
-int authenticate_user_krb5pwd(const char *user, const char *pswd, const char *service, const char *default_realm)
+int authenticate_user_krb5pwd(const char *user, const char *pswd, const char *service, const char *default_realm, unsigned char verify)
 {
     krb5_context    kcontext = NULL;
     krb5_error_code code;
@@ -38,59 +38,56 @@ int authenticate_user_krb5pwd(const char *user, const char *pswd, const char *se
     char            *name = NULL;
     char            *p = NULL;
 
+    // create kerberos context
     code = krb5_init_context(&kcontext);
-    if (code)
-    {
-        PyErr_SetObject(BasicAuthException_class, Py_BuildValue("((s:i))",
-                                                                "Cannot initialize Kerberos5 context", code));
+    if (code) {
+        PyErr_SetObject(BasicAuthException_class, Py_BuildValue("((s:i))", "Cannot initialize Kerberos5 context", code));
         return 0;
     }
 
-    ret = krb5_parse_name (kcontext, service, &server);
-
-    if (ret)
-    {
+    // parse service name to get the server principal
+    ret = krb5_parse_name(kcontext, service, &server);
+    if (ret) {
         set_basicauth_error(kcontext, ret);
         ret = 0;
         goto end;
     }
 
+    // unparse server pinrcipal again to get cannonical string representation
     code = krb5_unparse_name(kcontext, server, &name);
-    if (code)
-    {
+    if (code) {
         set_basicauth_error(kcontext, code);
         ret = 0;
         goto end;
     }
+
 #ifdef PRINTFS
     printf("Using %s as server principal for password verification\n", name);
 #endif
+
+    // free cannonical server principal
     free(name);
     name = NULL;
 
+    // construct user principal string from username and default realm
     name = (char *)malloc(256);
     p = strchr(user, '@');
-    if (p == NULL)
-    {
+    if (p == NULL) {
         snprintf(name, 256, "%s@%s", user, default_realm);
-    }
-    else
-    {
+    } else {
         snprintf(name, 256, "%s", user);
     }
 
+    // parse it into principal structure
     code = krb5_parse_name(kcontext, name, &client);
-    if (code)
-    {
+    if (code) {
         set_basicauth_error(kcontext, code);
         ret = 0;
         goto end;
     }
 
-    code = verify_krb5_user(kcontext, client, pswd, server);
-
-    if (code)
-    {
+    code = verify_krb5_user(kcontext, client, pswd, server, verify);
+    if (code) {
         ret = 0;
         goto end;
     }
@@ -101,6 +98,7 @@ end:
 #ifdef PRINTFS
     printf("kerb_authenticate_user_krb5pwd ret=%d user=%s authtype=%s\n", ret, user, "Basic");
 #endif
+
     if (name)
         free(name);
     if (client)
@@ -113,35 +111,61 @@ end:
 }
 
 /* Inspired by krb5_verify_user from Heimdal */
-static krb5_error_code verify_krb5_user(krb5_context context, krb5_principal principal, const char *password, krb5_principal server)
+static krb5_error_code verify_krb5_user(krb5_context context, krb5_principal principal, const char *password, krb5_principal server, unsigned char verify)
 {
     krb5_creds creds;
-    krb5_get_init_creds_opt gic_options;
+    krb5_get_init_creds_opt *gic_options;
+    krb5_verify_init_creds_opt vic_options;
     krb5_error_code ret;
     char *name = NULL;
 
     memset(&creds, 0, sizeof(creds));
 
+    // verify passed in client principal
     ret = krb5_unparse_name(context, principal, &name);
-    if (ret == 0)
-    {
+    if (ret == 0) {
 #ifdef PRINTFS
         printf("Trying to get TGT for user %s\n", name);
 #endif
         free(name);
     }
 
-    krb5_get_init_creds_opt_init(&gic_options);
-    ret = krb5_get_init_creds_password(context, &creds, principal, (char *)password, NULL, NULL, 0, NULL, &gic_options);
-    if (ret)
-    {
+    // verify passed in server principal if needed
+    if (verify) {
+        ret = krb5_unparse_name(context, server, &name);
+        if (ret == 0) {
+#ifdef PRINTFS
+            printf("Trying to get TGT for service %s\n", name);
+#endif
+            free(name);
+        }
+    }
+
+    // verify password
+    krb5_get_init_creds_opt_alloc(context, &gic_options);
+    ret = krb5_get_init_creds_password(context, &creds, principal, (char *)password, NULL, NULL, 0, NULL, gic_options);
+    if (ret) {
         set_basicauth_error(context, ret);
         goto end;
     }
 
+    // verify response authenticity
+    if (verify) {
+        krb5_verify_init_creds_opt_init(&vic_options);
+        krb5_verify_init_creds_opt_set_ap_req_nofail(&vic_options, 1);
+        ret = krb5_verify_init_creds(context, &creds, server, NULL, NULL, &vic_options);
+        if (ret) {
+            set_basicauth_error(context, ret);
+        }
+    }
+
 end:
+    // clean up
     krb5_free_cred_contents(context, &creds);
 
+    if (gic_options)
+        krb5_get_init_creds_opt_free(context, gic_options);
+
     return ret;
 }
 
